@@ -48,11 +48,10 @@ use constant STACK_ARGS         => 1;
 use constant STACK_BUFFER       => 2;
 use constant STACK_MODS         => 3;
 use constant STACK_PATH         => 4;
-use constant STACK_DEPTH        => 5;
-use constant STACK_BASE_COMP    => 6;
-use constant STACK_IN_CALL_SELF => 7;
-use constant STACK_BUFFER_IS_FLUSHABLE => 8;
-use constant STACK_HIDDEN_BUFFER => 9;
+use constant STACK_BASE_COMP    => 5;
+use constant STACK_IN_CALL_SELF => 6;
+use constant STACK_BUFFER_IS_FLUSHABLE => 7;
+use constant STACK_HIDDEN_BUFFER => 8;
 
 # HTML::Mason::Exceptions always exports rethrow_exception() and isa_mason_exception()
 use HTML::Mason::Exceptions( abbr => [qw(error param_error syntax_error
@@ -113,6 +112,10 @@ BEGIN
            regex => qr/^(?:output|fatal)$/,
            descr => "How error conditions are manifest (output or fatal)" },
 
+         component_error_handler =>
+         { parse => 'code', type => CODEREF|SCALAR, default => \&rethrow_exception,
+           descr => "A subroutine reference called on component compilation or runtime errors" },
+
          max_recurse =>
          { parse => 'string', default => 32, type => SCALAR,
            descr => "The maximum recursion depth for component, inheritance, and request stack" },
@@ -146,6 +149,7 @@ BEGIN
 my @read_write_params;
 BEGIN { @read_write_params = qw(
                                 autoflush
+                                component_error_handler
                                 data_cache_api
                                 data_cache_defaults
                                 dhandler_name
@@ -220,7 +224,8 @@ my %plugin_loaded;
 sub _initialize {
     my ($self) = @_;
 
-    local $SIG{'__DIE__'} = \&rethrow_exception;
+    local $SIG{'__DIE__'} = $self->component_error_handler
+        if $self->component_error_handler;
 
     eval {
         # Check the static_source touch file, if it exists, before the
@@ -375,8 +380,8 @@ sub exec {
     # at the bottom of _initialize(); just return.
     return unless $self->initialized();
 
-    # All errors returned from this routine will be in exception form.
-    local $SIG{'__DIE__'} = \&rethrow_exception;
+    local $SIG{'__DIE__'} = $self->component_error_handler
+        if $self->component_error_handler;
 
     # Cheap way to prevent users from executing the same request twice.
     #
@@ -981,9 +986,7 @@ sub decline
 #
 sub depth
 {
-    # direct access for speed because this method is called on every
-    # call to $m->comp
-    return $_[0]->{top_stack}->[STACK_DEPTH];
+    return scalar @{ $_[0]->{stack} };
 }
 
 #
@@ -1201,35 +1204,33 @@ sub comp {
         $comp = $self->fetch_comp($path, undef, \$error)
             or error($error || "could not find component for path '$path'\n");
     }
-    
+
     # Increment depth and check for maximum recursion. Depth starts at 1.
     #
-    my $top_stack = $self->{top_stack};
-    my $depth = defined($top_stack) ? $top_stack->[STACK_DEPTH] + 1 : 1;
-    error ($depth-1 . " levels deep in component stack (infinite recursive call?)\n")
-        if ($depth > $self->{max_recurse});
+    my $depth = $self->depth;
+    error "$depth levels deep in component stack (infinite recursive call?)\n"
+        if $depth >= $self->{max_recurse};
 
     # Keep the same output buffer unless store modifier was passed. If we have
     # a filter, put the filter buffer on the stack instead of the regular buffer.
     #
     my $filter_buffer = '';
-    my $top_buffer = defined($mods{store}) ? $mods{store} : $top_stack->[STACK_BUFFER];
+    my $top_buffer = defined($mods{store}) ? $mods{store} : $self->{top_stack}->[STACK_BUFFER];
     my $stack_buffer = $comp->{has_filter} ? \$filter_buffer : $top_buffer;
     my $flushable = exists $mods{flushable} ? $mods{flushable} : 1;
 
     # Add new stack frame and point dynamically scoped $self->{top_stack} at it.
-    local $self->{top_stack} = $self->{stack}->[$depth-1] =
-        [
-         $comp,           # STACK_COMP
-         \@_,             # STACK_ARGS
-         $stack_buffer,   # STACK_BUFFER
-         \%mods,          # STACK_MODS
-         $path,           # STACK_PATH
-         $depth,          # STACK_DEPTH
-         undef,           # STACK_BASE_COMP
-         undef,           # STACK_IN_CALL_SELF
-         $flushable       # STACK_BUFFER_IS_FLUSHABLE
-         ];
+    push @{ $self->{stack} },
+        [ $comp,           # STACK_COMP
+          \@_,             # STACK_ARGS
+          $stack_buffer,   # STACK_BUFFER
+          \%mods,          # STACK_MODS
+          $path,           # STACK_PATH
+          undef,           # STACK_BASE_COMP
+          undef,           # STACK_IN_CALL_SELF
+          $flushable,      # STACK_BUFFER_IS_FLUSHABLE
+        ];
+    local $self->{top_stack} = $self->{stack}->[-1];
 
     # Run start_component hooks for each plugin.
     #
@@ -1237,7 +1238,7 @@ sub comp {
         my $context = bless
             [$self, $comp, \@_],
             'HTML::Mason::Plugin::Context::StartComponent';
-        
+
         foreach my $plugin_instance (@{$self->{plugin_instances}}) {
             $plugin_instance->start_component_hook( $context );
         }
@@ -1249,12 +1250,17 @@ sub comp {
     my @result;
     
     eval {
-        if ($wantarray) {
-            @result = $comp->run(@_);
-        } elsif (defined $wantarray) {
-            $result[0] = $comp->run(@_);
-        } else {
-            $comp->run(@_);
+        # By putting an empty block here, we protect against stack
+        # corruption when a component calls next or last outside of a
+        # loop. See 05-request.t #28 for a test.
+        {
+            if ($wantarray) {
+                @result = $comp->run(@_);
+            } elsif (defined $wantarray) {
+                $result[0] = $comp->run(@_);
+            } else {
+                $comp->run(@_);
+            }
         }
     };
     my $error = $@;
@@ -1318,16 +1324,21 @@ sub content {
     # Run the content routine with the previous stack frame active and
     # with output going to a new buffer.
     #
+    my $err;
     my $buffer;
-    my $save_frame = $self->{top_stack};
+    my $save_frame = pop @{ $self->{stack} };
     {
-        local $self->{top_stack} = $self->_stack_frame(1);
+        local $self->{top_stack} = $self->{stack}[-1];
         local $self->{top_stack}->[STACK_BUFFER] = \$buffer;
         local $self->{top_stack}->[STACK_BUFFER_IS_FLUSHABLE] = 0;
         local $self->{top_stack}->[STACK_HIDDEN_BUFFER] = $save_frame->[STACK_BUFFER];
-        $content->();
+        eval { $content->(); };
+        $err = $@;
     }
-    $self->{top_stack} = $save_frame;
+
+    push @{ $self->{stack} }, $save_frame;
+
+    rethrow_exception $err;
 
     # Return the output from the content routine.
     #
@@ -1424,7 +1435,7 @@ sub debug_hook
 # 
 sub _stack_frame {
     my ($self, $levels) = @_;
-    my $depth = $self->{top_stack}->[STACK_DEPTH];
+    my $depth = $self->depth();
     my $index;
     if ($levels < 0) {
         $index = (-1 * $levels) - 1;
@@ -1439,7 +1450,8 @@ sub _stack_frame {
 # initial frame.
 sub _stack_frames {
     my ($self) = @_;
-    my $depth = $self->{top_stack}->[STACK_DEPTH];
+
+    my $depth = $self->depth;
     return reverse map { $self->{stack}->[$_] } (0..$depth-1);
 }
 
@@ -1455,7 +1467,7 @@ sub base_comp {
     return unless $self->{top_stack};
 
     unless ( defined $self->{top_stack}->[STACK_BASE_COMP] ) {
-        $self->_compute_base_comp_for_frame($self->{top_stack}->[STACK_DEPTH] - 1);
+        $self->_compute_base_comp_for_frame( $self->depth - 1 );
     }
     return $self->{top_stack}->[STACK_BASE_COMP];
 }
@@ -1661,6 +1673,19 @@ L<CGI|HTML::Mason::CGIHandler> is I<output>, causing the error to be
 displayed in the browser.  The default for standalone mode is
 I<fatal>.
 
+=item component_error_handler
+
+A code reference used to handle errors thrown during component
+compilation or runtime. By default, this is a subroutine that turns
+non-exception object errors in components into exceptions. If this
+parameter is set to a false value, these errors are simply rethrown
+as-is.
+
+Turning exceptions into objects can be expensive, since this will
+cause the generation of a stack trace for each error. If you are using
+strings or unblessed references as exceptions in your code, you may
+want to turn this off as a performance boost.
+
 =item max_recurse
 
 The maximum recursion depth for the component stack, for the request
